@@ -44,6 +44,7 @@ enum VectorType {
   VECTOR_TYPE_BLOB,
   VECTOR_TYPE_LIST_OF,
   VECTOR_TYPE_DATA_FRAME,
+  VECTOR_TYPE_MATRIX,
   VECTOR_TYPE_OTHER
 };
 
@@ -109,7 +110,7 @@ struct RConverter {
 
 static inline void warn_lossy_conversion(int64_t count, const char* msg) {
   SEXP fun = PROTECT(Rf_install("warn_lossy_conversion"));
-  SEXP count_sexp = PROTECT(Rf_ScalarReal(count));
+  SEXP count_sexp = PROTECT(Rf_ScalarReal((double)count));
   SEXP msg_sexp = PROTECT(Rf_mkString(msg));
   SEXP call = PROTECT(Rf_lang3(fun, count_sexp, msg_sexp));
   Rf_eval(call, nanoarrow_ns_pkg);
