@@ -85,6 +85,10 @@ test_that("infer_nanoarrow_schema() methods work for hms type", {
   expect_identical(infer_nanoarrow_schema(hms::hms())$format, "ttm")
 })
 
+test_that("infer_nanoarrow_schema() method works for matrix type", {
+  expect_identical(infer_nanoarrow_schema(matrix(ncol = 3))$format, "+w:3")
+})
+
 test_that("infer_nanoarrow_schema() methods work for vctrs types", {
   skip_if_not_installed("vctrs")
 
@@ -100,6 +104,25 @@ test_that("infer_nanoarrow_schema() method works for integer64()", {
   expect_identical(infer_nanoarrow_schema(bit64::integer64())$format, "l")
 })
 
+test_that("infer_nanoarrow_schema() method works for AsIs", {
+  expect_identical(
+    infer_nanoarrow_schema(I(integer()))$format,
+    infer_nanoarrow_schema(integer())$format
+  )
+})
+
+test_that("infer_nanoarrow_schema() returns list of null for empty or all null list", {
+  expect_identical(infer_nanoarrow_schema(list())$format, "+l")
+  expect_identical(infer_nanoarrow_schema(list())$children[[1]]$format, "n")
+  expect_identical(infer_nanoarrow_schema(list(NULL))$format, "+l")
+  expect_identical(infer_nanoarrow_schema(list())$children[[1]]$format, "n")
+})
+
+test_that("infer_nanoarrow_schema() returns binary for list of raw", {
+  expect_identical(infer_nanoarrow_schema(list(raw()))$format, "z")
+  expect_identical(infer_nanoarrow_schema(list(raw(), NULL))$format, "z")
+})
+
 test_that("nanoarrow_schema_parse() works", {
   simple_info <- nanoarrow_schema_parse(na_int32())
   expect_identical(simple_info$type, "int32")
@@ -261,9 +284,10 @@ test_that("schema modify can modify name", {
 
 test_that("schema modify can modify flags", {
   schema <- na_int32()
+  expect_identical(schema$flags, 2L)
 
-  schema2 <- nanoarrow_schema_modify(schema, list(flags = 1))
-  expect_identical(schema2$flags, 1L)
+  schema2 <- nanoarrow_schema_modify(schema, list(flags = 0))
+  expect_identical(schema2$flags, 0L)
   expect_identical(schema2$format, schema$format)
   expect_identical(schema2$name, schema$name)
 
